<?php

namespace Myriad\EventDrivenService\Services;

use App\Exceptions\ApiBadRequestException;
use Myriad\EventDrivenService\Http\Utilities\EventBuilder;
use Myriad\EventDrivenService\Http\Utilities\EventHelper;
use Myriad\EventDrivenService\Http\Utilities\Pagination;
use Myriad\EventDrivenService\Models\EventModel;
use Myriad\EventDrivenService\Repositories\EventRepository;
use Myriad\EventDrivenService\Repositories\FailedEventRepository;
use Myriad\EventDrivenService\Traits\Logger;
use Illuminate\Http\Request;
use Illuminate\Validation\ValidationException;

class EventService {

  use Logger;

  private $eventRepository;
  private $failedEventRepository;

  public function __construct(EventRepository $eventRepository, FailedEventRepository $failedEventRepository) {
    $this->eventRepository = $eventRepository;
    $this->failedEventRepository = $failedEventRepository;
  }

  public function getNew($uri, $lastUpdate, $pageSize) {
    $model = EventHelper::getModelByUri($uri);

    $recordsFound = $model->getUpdated($lastUpdate, $pageSize);

    $currentData = [];
    foreach ($recordsFound as $record) {
      $currentData[] = EventBuilder::make(EventHelper::getCommandForModel($model), $record);
    }

    return [
      'data' => collect($currentData),
      'meta' => Pagination::build($recordsFound)
    ];
  }

  public function handlePost(Request $request) {
    list($valid, $invalid) = collect($request->data)
      ->partition(function ($value, $key) {
        return isset($value['command']) && isset($value['model_uuid']) ? $value : null;
      });

    $valid = $valid->groupBy([function ($key, $value) {
      list($model, $command) = explode(':', $key['command']);
      return $model;
    }, 'model_uuid']);

    $errors = $invalid->map(function ($value, $key) {
      return $this->logError($value, "Model UUID and/or Command not provided.");
    });

    foreach ($valid as $modelType => $uuids) {
      $model = EventHelper::getModelByType($modelType);
      foreach ($uuids as $uuid => $events) {
        $currentModel = $model->withTrashed()->find($uuid);
        $created = false;
        if (!$currentModel) {
          $created = true;
          $currentModel = $model->create(['uuid' => $uuid])->refresh();
        }
        if ($currentModel->trashed()) {
          $deletedEvents = collect($events)->map(function($event) {
            return [
              'request' => $event,
              'error' => 'Model is deleted.'
            ];
          });
          $errors = $errors->merge($deletedEvents);
          continue;
        }
        foreach ($events as $event) {
          list($event['model'], $event['operation']) = explode(':', $event['command']);
          try {
            \Validator::make($event, [
              'operation' => 'required|in:' . implode($currentModel->getCommands(), ",")
            ])->validate();

            if ($event['operation'] == 'create' && $created) {
              throw new ApiBadRequestException('Model already created.');
            }

            $response = call_user_func_array([$this, $event['operation']], [$currentModel, $event]);
            if ($response) {
              $errors[] = $response;
            }
          } catch (ValidationException $e) {
            $errors[] = $this->logValidationError($event, $e);
            continue;
          } catch (\Exception $e) {
            $errors[] = $this->logError($event, $e->getMessage());
            continue;
          }
        }
      }
    }
    $this->failedEventRepository->createAll($errors);
    return $errors;
  }

  public function create(EventModel $model, $event) {
    try {
      \Validator::make($event, [
        'value' => 'required',
        'model_uuid' => 'required'
      ])->validate();
    } catch (ValidationException $e) {
      throw $e;
    }

    try {
      \DB::beginTransaction();
      $this->eventRepository->create($event);
      $model->createModel($event);
      \DB::commit();
    } catch (\Exception $e) {
      \DB::rollBack();
      throw $e;
    }
  }

  public function update(EventModel $model, $event) {
    try {
      \Validator::make($event, [
        'model_uuid' => 'required|string',
        'name' => 'required|string',
        'value' => 'required',
        'version' => 'string|nullable',
        'version_new' => 'required|string'
      ])->validate();
    } catch (ValidationException $e) {
      throw $e;
    }

    if (!array_key_exists($event['name'], $model->getAttributes())) {
      throw new ApiBadRequestException("Column not found");
    }

    if (!array_key_exists($event['name'] . '_version', $model->getAttributes())) {
      throw new ApiBadRequestException("Column is not an offline column");
    }

    if (!$model->isCurrent($event['name'], $event['version'])) {
      throw new ApiBadRequestException("Newer version already exists");
    }

    try {
      \DB::beginTransaction();
      $this->eventRepository->create($event);
      $model->updateModel($event);
      \DB::commit();
    } catch (\Exception $e) {
      \DB::rollBack();
      throw $e;
    }
  }

  public function delete(EventModel $model, $event) {
    try {
      \DB::beginTransaction();
      $this->eventRepository->create($event);
      $model->deleteModel($event);
      \DB::commit();
    } catch (\Exception $e) {
      \DB::rollBack();
      throw $e;
    }
  }
}